"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const nodeDescription_1 = require("../interfaces/nodeDescription");
const nodeType_1 = require("../interfaces/nodeType");
const TypedArray = Object.getPrototypeOf(Int8Array);
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function buildDescriptionTreeRecursive(nodeKey, node, level, settings) {
    const currentLevelMaxItems = (level >= settings.maxLevel)
        ? 0
        : Array.isArray(settings.maxItemsPerLevel)
            ? settings.maxItemsPerLevel[Math.min(level, settings.maxItemsPerLevel.length - 1)]
            : settings.maxItemsPerLevel;
    let nodeDescription;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    let unwrappedSubEntries = [];
    if (node === undefined) {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Undefined],
            value: 'undefined',
        };
    }
    else if (node === null) {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Null],
            value: 'null',
        };
    }
    else if (typeof node === 'string') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.String],
            value: node,
        };
    }
    else if (typeof node === 'number') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Number],
            value: parseFloat(node.toPrecision(6)).toLocaleString(),
        };
    }
    else if (typeof node === 'bigint') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.BigInt],
            value: node.toLocaleString(),
        };
    }
    else if (typeof node === 'boolean') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Boolean],
            value: node.toLocaleString(),
        };
    }
    else if (typeof node === 'symbol') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Symbol],
            value: node.toLocaleString(),
        };
    }
    else if (typeof node === 'function') {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Function],
            value: node.name,
        };
    }
    else if (node && typeof node === 'object') {
        let nodeCustomObjectString;
        if (node.toLocaleString) {
            const jsObjectString = node.toLocaleString();
            if (jsObjectString !== '[object Object]') {
                nodeCustomObjectString = jsObjectString;
            }
        }
        if (node instanceof String) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.StringObject],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof Number) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.NumberObject],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof Boolean) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.BooleanObject],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof Date) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Date],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof WeakMap) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.WeakMap],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof WeakSet) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.WeakMap],
                value: node.toLocaleString(),
            };
        }
        else if (node instanceof Error) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Enumerable, nodeType_1.EnumerableNodeFineTypeEnum.Error],
                value: node.toLocaleString(),
                subEntries: [],
            };
            unwrappedSubEntries = Object.entries(node);
            unwrappedSubEntries.push(['stack', node.stack]); // TODO: whitelist\blacklist of special unenumerable props
        }
        else if (node instanceof Map) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Enumerable, nodeType_1.EnumerableNodeFineTypeEnum.Map],
                value: node.toLocaleString(),
                subEntries: [],
            };
            unwrappedSubEntries = Array.from(node.entries());
        }
        else if (node instanceof Set) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Iterable, nodeType_1.IterableNodeFineTypeEnum.Set],
                value: node.toLocaleString(),
                subEntries: [],
            };
            unwrappedSubEntries = Array.from(node.entries());
        }
        else if (Array.isArray(node)) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Iterable, nodeType_1.IterableNodeFineTypeEnum.Array],
                value: '',
                subEntries: [],
            };
            unwrappedSubEntries = Object.entries(node);
        }
        else if (node instanceof TypedArray) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Iterable, nodeType_1.IterableNodeFineTypeEnum.TypedArray],
                value: node.toLocaleString(),
                subEntries: [],
            };
            const typedArray = node;
            const array = Array.from(typedArray);
            unwrappedSubEntries = Object.entries(array);
        }
        else if (node[Symbol.iterator]) {
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Iterable, nodeType_1.IterableNodeFineTypeEnum.Iterable],
                value: node.toLocaleString(),
                subEntries: [],
            };
            const array = Array.from(node);
            unwrappedSubEntries = Object.entries(array);
        }
        else if (node.hasOwnProperty) {
            let nodeValue;
            if (nodeCustomObjectString) {
                nodeValue = nodeCustomObjectString;
            }
            else {
                nodeValue = ''; // TODO!!
            }
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Enumerable, nodeType_1.EnumerableNodeFineTypeEnum.Object],
                subEntries: [],
                value: nodeValue,
            };
            unwrappedSubEntries = Object.entries(node);
        }
        else {
            let nodeValue;
            if (nodeCustomObjectString) {
                nodeValue = nodeCustomObjectString;
            }
            else {
                nodeValue = 'DETAILLESS OBJECT';
            }
            nodeDescription = {
                key: nodeKey,
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Enumerable, nodeType_1.EnumerableNodeFineTypeEnum.Unknown],
                value: nodeValue,
                subEntries: [],
            };
            unwrappedSubEntries = [];
        }
    }
    else {
        nodeDescription = {
            key: nodeKey,
            typeTuple: [nodeType_1.NodeBroadTypeEnum.Single, nodeType_1.SingleNodeFineTypeEnum.Unknown],
            value: (typeof node).toLocaleUpperCase(),
        };
    }
    if (nodeDescription_1.guardNodeDescription(nodeType_1.NodeBroadTypeEnum.Enumerable, nodeDescription) ||
        nodeDescription_1.guardNodeDescription(nodeType_1.NodeBroadTypeEnum.Iterable, nodeDescription)) {
        const itemsShownCount = Math.min(currentLevelMaxItems, unwrappedSubEntries.length);
        for (let i = 0; i < itemsShownCount; i++) {
            const [subEntryKey, subEntry] = unwrappedSubEntries[i];
            // TODO limits ??
            const childDesctiption = buildDescriptionTreeRecursive(subEntryKey, subEntry, level + 1, settings);
            nodeDescription.subEntries.push(childDesctiption);
        }
        if (nodeDescription.subEntries.length < unwrappedSubEntries.length) {
            const elipsisDescription = {
                typeTuple: [nodeType_1.NodeBroadTypeEnum.Dead, nodeType_1.DeadNodeFineTypeEnum.Elipsis],
            };
            nodeDescription.subEntries.push(elipsisDescription);
        }
        if (nodeDescription.subEntries.length < unwrappedSubEntries.length) {
            nodeDescription.info = `${nodeDescription.subEntries.length} shown of ${unwrappedSubEntries.length} total`;
        }
        else {
            nodeDescription.info = `${nodeDescription.subEntries.length} entries`;
        }
    }
    return nodeDescription;
}
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function buildDescriptionTree(nodeKey, node, settings) {
    return buildDescriptionTreeRecursive(nodeKey, node, 0, settings);
}
exports.buildDescriptionTree = buildDescriptionTree;
//# sourceMappingURL=buildDescriptionTree.js.map