"use strict";
/* eslint-disable @typescript-eslint/ban-types */
Object.defineProperty(exports, "__esModule", { value: true });
const eolRegexp = /[\n\r]/g;
class NormalizedUnicodeText extends String {
    constructor(text, isSkipChecks = false) {
        const isNormalizedInstance = NormalizedUnicodeText.isNormalizedInstance(text);
        if (isSkipChecks || isNormalizedInstance) {
            super(text);
        }
        else {
            const normalizedText = text.normalize();
            super(normalizedText);
        }
    }
    normalize() {
        return this.valueOf();
    }
    static isNormalizedInstance(text) {
        return text instanceof NormalizedUnicodeText;
    }
    static combine(...items) {
        const combinedString = items.reduce((combinedString, item) => {
            return combinedString + item.toString();
        }, '');
        return new NormalizedUnicodeText(combinedString);
    }
}
exports.NormalizedUnicodeText = NormalizedUnicodeText;
class StrictUnicodeText extends NormalizedUnicodeText {
    constructor(text, isSkipChecks = false) {
        super(text, isSkipChecks);
        isSkipChecks = isSkipChecks || this.isPrecheckedInstance(text);
        if (!isSkipChecks) {
            this.guardStringIsStrictUnicode(this);
        }
    }
    isPrecheckedInstance(text) {
        return text instanceof StrictUnicodeText;
    }
    guardStringIsStrictUnicode(normalizedText) {
        for (const utf16Code of normalizedText) {
            const utf16CodePoint = utf16Code.codePointAt(0);
            if (utf16CodePoint === undefined || utf16CodePoint > 65536) {
                // TODO: 1. ASAP check limits
                // TODO: 2. later support compound multi utf16Code
                throw new Error('String is nonstrict unicode, which is unsupported');
            }
        }
    }
    splitToFeedLines() {
        const feedLines = this.valueOf().split(eolRegexp).map((lineString) => {
            return new StrictUnicodeLine(lineString, true);
        });
        return feedLines;
    }
    static combine(...items) {
        const combinedNormalized = super.combine(...items);
        return new StrictUnicodeText(combinedNormalized);
    }
    *managedWrap(maxWidth) {
        const feedLines = this.splitToFeedLines();
        let wrappedLine;
        for (let lineId = 0; lineId < feedLines.length; lineId++) {
            const feelLine = feedLines[lineId];
            const isLastLine = lineId === feedLines.length - 1;
            const lineWrapGenerator = feelLine.managedWrap(maxWidth);
            let done;
            do {
                const generatorResult = lineWrapGenerator.next(maxWidth);
                done = generatorResult.done || false;
                wrappedLine = generatorResult.value;
                if (!done || !isLastLine) {
                    maxWidth = yield wrappedLine;
                }
            } while (!done);
        }
        // tail
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        return wrappedLine;
    }
}
exports.StrictUnicodeText = StrictUnicodeText;
class StrictUnicodeLine extends StrictUnicodeText {
    constructor(text, isSkipChecks = false) {
        if (eolRegexp.test(text.toString())) {
            console.log(text.toString());
        }
        super(text, isSkipChecks);
    }
    isPrecheckedInstance(text) {
        return text instanceof StrictUnicodeLine;
    }
    guardStringIsStrictUnicode(normalizedText) {
        if (eolRegexp.test(normalizedText.toString())) {
            throw ('No EOLs allowed in single StrictUnicodeLine, use StrictUnicodeText instead');
        }
        super.guardStringIsStrictUnicode(normalizedText);
    }
    calcWidth() {
        if (this.widthCache === undefined) {
            const iterator = this[Symbol.iterator]();
            let width = 0;
            while (!iterator.next().done) {
                width++;
            }
            this.widthCache = width;
            // or
            // for ({} of this) { width++ }
            // or
            // [...this].length
        }
        return this.widthCache;
    }
    static combine(...items) {
        const combinedText = super.combine(...items);
        return new StrictUnicodeLine(combinedText);
    }
    *managedWrap(maxWidth) {
        const lineWidth = this.calcWidth();
        if (lineWidth <= maxWidth) {
            return this;
        }
        let wrappedLineWidth = 0;
        let wrappedLine = '';
        for (const codePoint of this) {
            const codePointWidth = 1; // TODO: support for full-width chars
            if (wrappedLineWidth + codePointWidth > maxWidth) {
                maxWidth = yield new StrictUnicodeLine(wrappedLine);
                wrappedLine = '';
                wrappedLineWidth = 0;
            }
            wrappedLineWidth += codePointWidth;
            wrappedLine += codePoint;
        }
        //tail
        return new StrictUnicodeLine(wrappedLine);
    }
}
exports.StrictUnicodeLine = StrictUnicodeLine;
class StrictUnicodeChar extends StrictUnicodeLine {
    constructor() {
        super(...arguments);
        this.widthCache = 1;
    }
    isPrecheckedInstance(text) {
        return text instanceof StrictUnicodeChar;
    }
    guardStringIsStrictUnicode(normalizedText) {
        if (normalizedText.calcWidth() !== 1) {
            throw ('char width must be strictly 1');
        }
        super.guardStringIsStrictUnicode(normalizedText);
    }
    calcWidth() {
        return 1;
    }
    static combine(...items) {
        return super.combine(...items);
    }
}
exports.StrictUnicodeChar = StrictUnicodeChar;
//# sourceMappingURL=strictUnicode.js.map